/* 
    Autogenerated using:
    - Template: 'include/cubnm/models/model.cuh.mako'
    - Model definition: 'codegen/recipes/wc.yaml'
    Do not modify this autogenerated code. Instead modify template
    or model definition.
*/
#ifndef WC_CUH
#define WC_CUH
#include "wc.hpp"

// Device constant memory for model constants
__constant__ WCModel::Constants d_WCc;

// Explicit template instantiations for GPU kernels
// These are necessary because CUDA code is compiled separately by nvcc
// but the functions are called from the main code compiled by g++
// Therefore, from nvcc's perspective, there is no need to compile the
// function for this model unless we explicitly instantiate the templates
template void _run_simulations_gpu<WCModel>(
    double*, double*, double*, 
    double**, double**, double*, 
    double**, int*, double*, 
    BaseModel*
);

template void _init_gpu<WCModel>(BaseModel*, BWConstants, bool);
#endif