# Autogenerated based on 'src/cubnm/sim/__init__.py.mako'
# and list of available models.
# Do not modify this autogenerated code. Instead modify template.
from .base import SimGroup, create_multi_sim_group

from .jr import JRSimGroup
from .kuramoto import KuramotoSimGroup
from .rww import rWWSimGroup
from .rwwex import rWWExSimGroup
from .wc import WCSimGroup

__all__ = [
    'SimGroup',
    'JRSimGroup',
    'KuramotoSimGroup',
    'rWWSimGroup',
    'rWWExSimGroup',
    'WCSimGroup',
]